;//
;//	"FLATQUAD.ASM" - Sample Code
;//
;//	Copyright (C) 1998  Sony Computer Entertainment Europe
;//	All rights Reserved
;//

;// This file contains an HMD driver for flat quadrangles. Screen clipping is
;// performed.
;// Assemble with ASMPSX.

;// Based on the original library driver
;// Assumes the 'offset' in the GsOT structure is zero

SCREEN_HEIGHT equ 256
SCREEN_WIDTH  equ 512

BITMASK equs "s0"
CLIP_Y 	equs "s1"
CLIP_X 	equs "s2"
PARAM	equs "s3"
SCREEN_X equs "s3"
SCREEN_Y equs "s4"
FLIPSTATUS equs "s5"

	include gtereg.h
	include inline_a.h

	XDEF	GsU_00040051

	XREF	GsOUT_PACKET_P
	XREF	reflectionMode

	SECTION	.text

	OPT at-,h-

************************************************************************
* (Back-Facing) Flat/Texture Quad with no lighting
*
*	a0: (GsARGUNIT *)sp
*
*	typedef struct {
*		u_long	*primp;		// 0(a0) Data count/size of this primitive
*		GsOT 	*tagp;		// 4(a0)
*		int	shift;		// 8(a0)
*		int	offset;		// 12(a0)
*		PACKET	*out_packetp;	// 16(a0)
*		u_long	*primtop;	// 20(a0)	// PrimHeader top
*		SVECTOR	*vertop;	// 24(a0)	// Vertex top
*		SVECTOR	*nortop;	// 28(a0)	// Normal top
*	}	GsARGUNIT;
*	num:				// (*primtop) >> 16
*************************************************************************/

GsU_00040051:
		;// Prologue - we need some more temporary registers, so we have to
		;// save the s0-sX range.
		;// Not exactly a true prologue, but this is a leaf function any-hoo.
	subu sp,7*4
	sw	s0, 0(sp)	// Save
	sw	s1, 8(sp)
	sw	s2,12(sp)
	sw	s3,16(sp)
	sw	s4,20(sp)
	sw	s5,24(sp)
	sw	ra,28(sp)
		;// End prologue

	li	CLIP_Y,-SCREEN_HEIGHT/2		;// Y clipping value
	li	CLIP_X,-SCREEN_WIDTH/2		;// X clipping value
	lw  FLIPSTATUS,reflectionMode   ;// 0 if we shouldn't 'flip the clip'

	lw	a3,0(a0)			;// a3: sp->primp
	lw	t1,4(a0)			;// t1: otp 	// Ordering table start address
	lw	t8,0(a3)			;// *(sp->primp)	// = Data count/size
	lw	t7,8(a0)			;// Ordering table shift
	addu a3,a3,4			;// skip Data count/size, gives us offset of prims
	lw	a2,0(a3)			;// prim idx: *(sp->primp)	// Offset (from pointer in
							;// Primitive Header) where the poly prims are
	lw	a1,20(a0)			;// sp->primtop  Primitive Header Top
	sll	a2,a2,2				;// idx*4
	srl	t8,t8,16			;// n = number to process

	addu a3,a1,a2			;// primtop: sp->primtop+idx	// Address of the poly prims

	lw	t0,4(t1)			;// t0: table
	addu v0,a0,16
	lw	t1,0(v0)			;// t1: pkt_base

	lw	a1,24(a0)			;// a1: vertop

		;// At this stage:
		;//		a3 = pointer to the poly primitives
		;//		t8 = number of primitives to process
		;//		t7 = OT shift
		;//		t0 = where the OTable is.
		;//		t1 = packet
		;//		a1 = vertex top

	lhu	t2,14(a3)			;// t2: primtop->v0
	lhu	t3,18(a3)			;// t2: primtop->v1
	lhu	t4,20(a3)			;// t3: primtop->v2

	sll	t2,t2,3				;// v0*8
	sll	t3,t3,3				;// v1*8
	sll	t4,t4,3				;// v1*8
	addu t2,t2,a1			;// t2: (vertop+primtop->v0)
	addu t3,t3,a1			;// t3: (vertop+primtop->v1)
	addu t4,t4,a1			;// t4: (vertop+primtop->v2)

NTF4_1:
	lwc2 C2_VXY0,(t2)		;//Vector 0 <--(vertop+primtop->v0)
	lwc2 C2_VZ0,4(t2)
	lwc2 C2_VXY1,(t3)		;//Vector 1 <--(vertop+primtop->v1)
	lwc2 C2_VZ1,4(t3)
	lwc2 C2_VXY2,(t4)		;//Vector 2 <--(vertop+primtop->v2)
	lwc2 C2_VZ2,4(t4)
	nop
	nop
	;//===================================================================
		RTPT;
	;//===================================================================

	lw	v0,0(a3)		;// v0: load primtop->rgb
	lhu	v1,22(a3)		;// v1: load primtop->v3
	beqz FLIPSTATUS,NODARKEN
	nop
	li	v0,0x2c202020	;32 all the way
NODARKEN:
	sw	v0,4(t1)		;// v0: save primtop->rgb
	sll	v1,v1,3			;// v1: v3*8

	addu	a3,a3,24		;// primtop++ (primtop += sizeof(HMD_POLY_FT4))

	addu	v1,v1,a1		;// v1: (vertop+primtop->v3)

	lhu	t2,14(a3)		;// t2: primtop->v0
	lhu	t3,18(a3)		;// t2: primtop->v1
	lhu	t4,20(a3)		;// t3: primtop->v2

	sll	t2,t2,3			;// v0*8
	sll	t3,t3,3			;// v1*8
	sll	t4,t4,3			;// v1*8
	addu	t2,t2,a1		;// t2: (vertop+primtop->v0)
	addu	t3,t3,a1		;// t3: (vertop+primtop->v1)
	addu	t4,t4,a1		;// t4: (vertop+primtop->v2)

	cfc2	v0,C2_FLAG		;// v0: flag
	nop
	blt	v0,0,NTF4_2		;// clip if flag <0
	nop

	;//===================================================================
	NCLIP
	;//===================================================================

	lwc2	C2_VXY0,(v1)		;//Vector 0 <--(vertop+primtop->v3)
	lwc2	C2_VZ0,4(v1)

	mfc2	v0,C2_MAC0		;// opz read
	beqz	FLIPSTATUS,FLIP_CLIP
	nop
	neg		v0
FLIP_CLIP:
	ble	v0,0,NTF4_2		;// Normal clip if MAC0(OPZ)<=0
	nop

	   ;	// Saving and CLIPPING!

	li		BITMASK,0		;// Mask to tell us if we should trivally reject or not.

	mfc2	PARAM,C2_SXY0
	jal		CLIP
	swc2    C2_SXY0,8(t1) 	;// sxy0 -->8(pkt_base)

	mfc2	PARAM,C2_SXY1
	jal		CLIP
	swc2    C2_SXY1,16(t1)	;// sxy1 -->16(pkt_base)

	mfc2	PARAM,C2_SXY2
	jal		CLIP
	swc2    C2_SXY2,24(t1)	;// sxy2 -->24(pkt_base)
	nop
	nop

	;//===================================================================
	RTPS
	;//===================================================================

	mfc2	PARAM,C2_SXY2
	jal		CLIP
	swc2    C2_SXY2,32(t1)	;// sxy2 -->32(pkt_base)

			;// If any of the 4 bit masks are 1, then there must be 4 points off screen, so clip
	and		BITMASK,0x4444
	bne		BITMASK,zero,NTF4_2

		;// Doesn't matter that this instruction is in the delay slot of
		;// the above branch
	subu v0,a3,24		;// v0: (primtop-24)=old primtop
	lw	v1,4(v0)		;// primtop->uv0
	lw	t6,8(v0)		;// primtop->uv1
	sw	v1,12(t1)		;// uv0 --> 12(pkt_base)
	sw	t6,20(t1)		;// uv1 --> 20(pkt_base)
	lw	v1,12(v0)		;// primtop->uv2
	lw	t6,16(v0)		;// primtop->uv3
	sw	v1,28(t1)		;// uv2 --> 28(pkt_base)
	sw	t6,36(t1)		;// uv3 --> 36(pkt_base)

	cfc2 v0,C2_FLAG		;// v0: flag
	nop
	blt	v0,0,NTF4_2		;// clip if flag <0
	nop


	;//===================================================================
	AVSZ4
	;//===================================================================

	mfc2	t6,C2_OTZ	;// t6: otz
	nop

	srl 	t6,t6,t7	;// sz = sz(otz) >> shift;
	and	t6,t6,0xffff	;// mask ?
	sll	t6,t6,2			;// sz*4
	addu	t6,t6,t0	;// t6: &(table[sz])
	lw	t5,0(t6)		;// t5: table[sz].p

	li	v0,0x9000000	;// (sizeof(POLY_FT4)/4-1)<<24)
	sll	v1,t5,8			;//
	srl	v1,v1,8			;// v1: clear (31..24) of table[sz].p
	or	v0,v0,v1		;// v0: (table[sz].p)|(sizeof(POLY_FT4)/4-1)<<24)
	sw	v0,0(t1) 		;// *(long *)pkt_base = v0

	xor	v0,t1,t5		;// v0 : pkt_base xor table[sz].p
	sll	v1,v0,8
	srl	v0,v1,8			;// v1: clear (31..24) of t3
	xor	v1,v0,t5		;// v1: v1 xor table[sz].p
	sw	v1,0(t6)		;// table[sz].p = pkt_base;

	addu	t1,t1,40	;// pkt_base++;  (pktBase += sizeof(POLY_FT4))

NTF4_2:

	subu	t8,t8,1		;//NumPrimstoProcess--;
	bne	t8,0,NTF4_1		;// if(primtop != primtop+n) goto NTF4_1
	nop

NTF4_3:
	lw	v0,0(a0)		;// sp->primp

	OPT at+
	sw	t1,GsOUT_PACKET_P
	OPT at-
	addu	v0,v0,8			;// sp->primp+2 = address of next HMD prim

		;//Epilogue - Restore the stack
	lw	s0, 0(sp)	// Restore saved args
	lw	s1, 8(sp)
	lw	s2,12(sp)
	lw	s3,16(sp)
	lw	s4,20(sp)
	lw	s5,24(sp)
	lw	ra,28(sp)
	addu	sp,7*4		; Delay slot
		;//End of Epilogue

	j       ra			;// return(primtop)
	nop

	opt at+
CLIP:
	sra		SCREEN_Y,PARAM,16			;// s4 = signed SY value;
	sll		SCREEN_X,PARAM,16
		;// Check Upper boundary
	bge		SCREEN_Y,CLIP_Y,@SY_OK_UP	;//jump if SCREEN_Y < -128
	sra		SCREEN_X,SCREEN_X,16		;// Sign extend X. (Delay slot)

	addu	BITMASK,BITMASK,0x0001
@SY_OK_UP:
		;// Check left boundary
	bge		SCREEN_X,CLIP_X,@SX_OK_LEFT
	neg		SCREEN_Y

	addu	BITMASK,BITMASK,0x0010
@SX_OK_LEFT:
		;// Check lower boundary
	bgt		SCREEN_Y,CLIP_Y,@SY_OK_DOWN
	neg		SCREEN_X

	addu	BITMASK,BITMASK,0x0100
@SY_OK_DOWN:
		;// Check right boundary
	bgt		SCREEN_X,CLIP_X,@SY_OK_RIGHT
	nop
	addu	BITMASK,BITMASK,0x1000
@SY_OK_RIGHT:
	j		ra
	nop
