;//
;//	"FLATQUAD.ASM" - Sample Code
;//
;//	Copyright (C) 1998  Sony Computer Entertainment Europe
;//	All rights Reserved
;//

;// This file contains an HMD driver for flat triangles. Screen clipping is
;// performed.
;// Assemble with ASMPSX.

;// Based on the original library driver
;// Assumes the 'offset' in the GsOT structure is zero

SCREEN_HEIGHT equ 256
SCREEN_WIDTH  equ 512

BITMASK equs "s0"
CLIP_Y 	equs "s1"
CLIP_X 	equs "s2"
PARAM	equs "s3"
SCREEN_X equs "s3"
SCREEN_Y equs "s4"
FLIPSTATUS equs "s5"

	include gtereg.h
	include inline_a.h

	XDEF	GsU_00040049

	XREF	GsOUT_PACKET_P
	XREF	reflectionMode

	SECTION	.text

	OPT	at-,h-

************************************************************************
* Back-Facing Flat/Textured Triangle with no lighting
*
*	a0: (GsARGUNIT *)sp
*
*	typedef struct {
*		u_long	*primp;		// 0(a0)
*		GsOT 	*tagp;		// 4(a0)
*		int	shift;		// 8(a0)
*		int	offset;		// 12(a0)
*		PACKET	*out_packetp;	// 16(a0)
*		u_long	*primtop;	// 20(a0)
*		SVECTOR	*vertop;	// 24(a0)
*		SVECTOR	*nortop;	// 28(a0)
*	}	GsARGUNIT;
*	num:				// (*primtop) >> 16
*************************************************************************/

GsU_00040049:
		;// Prologue - we need some more temporary registers, so we have to
		;// save the s0-sX range.
		;// Not exactly a true prologue, but this is a leaf function any-hoo.
	subu sp,7*4
	sw	s0, 0(sp)
	sw	s1, 8(sp)
	sw	s2,12(sp)
	sw	s3,16(sp)
	sw	s4,20(sp)
	sw	s5,24(sp)
	sw	ra,28(sp)
		;// End prologue

	li	CLIP_Y,-SCREEN_HEIGHT/2		;// Y clipping value
	li	CLIP_X,-SCREEN_WIDTH/2		;// X clipping value
	lw  FLIPSTATUS,reflectionMode

	lw	a3,0(a0)		;// a3: sp->primp
	lw	t1,4(a0)		;// sp->tagp
	lw	t8,0(a3)		;// *(sp->primp)
	lw	t7,8(a0)		;// sp->shift
	addu a3,a3,4			;// skip size, num
	lw	a2,0(a3)		;// prim idx:*(sp->primp)

	lw	a1,20(a0)		;// sp->primtop
	sll	a2,a2,2			;// idx*4
	srl	t8,t8,16		;// t8: num

	addu a3,a1,a2		;// primtop: sp->primtop+idx

	lw	t0,4(t1)		;// t0: otp->org
	addu v0,a0,16		;// t1: pkt_base
	lw	t1,0(v0)		;// t1: out_packetp

	lw	a1,24(a0)		;// a1: vertop

		;// At this stage:
		;//		a3 = pointer to the poly primitives
		;//		t8 = number of primitives to process
		;//		t7 = OT shift
		;//		t0 = where the OTable is.
		;//		t1 = packet
		;//		a1 = vertex top

	lw	t2,12(a3)				;// t2: primtop->v0
	lw	t3,16(a3)				;// t3: primtop->v1,v2
	srl	t4,t2,16				;// t4: v0

	sll	t4,t4,3					;// t4: v0*8

	addu	t4,t4,a1			;// t4: (vertop+primtop->v0)

TF3L_1:
	lwc2	C2_VXY0,(t4)		;//Vector 0 <--(vertop+primtop->v0)
	lwc2	C2_VZ0,4(t4)

	sll	t5,t3,16				;// t5: v1
	srl	t5,t5,13				;// t5: v1*8
	addu	t5,t5,a1			;// t5: (vertop+primtop->v1)
	lwc2	C2_VXY1,(t5)		;//Vector 1 <--(vertop+primtop->v1)
	lwc2	C2_VZ1,4(t5)

	nop
	srl	t5,t3,16				;// t5: v2
	sll	t5,t5,3					;// t5: v2*8
	addu	t5,t5,a1			;// t5: (vertop+primtop->v2)
	lwc2	C2_VXY2,(t5)		;//Vector 2 <--(vertop+primtop->v2)
	lwc2	C2_VZ2,4(t5)
	sll	t5,t2,16				;// t5: n0<<16
	srl	t5,t5,13				;// t5: n0*8

	;//===========================================================
	RTPT
	;//===========================================================

	lw	v0, 0(a3)				;// v0: primtop->rgb_code
	lw	t2, 4(a3)				;// t2: primtop->uv0
	lw	t3, 8(a3)				;// t3: primtop->uv1
	lw	t4,12(a3)				;// t4: primtop->uv2
	beqz FLIPSTATUS,NODARKEN
	nop
	li	v0,0x24202020			;// RGB = 32
NODARKEN:
	sw	v0, 4(t1)				;// v0: out_packetp->rgb_code
	sw	t2,12(t1)				;// t2: out_packetp->uv0
	sw	t3,20(t1)				;// t3: out_packetp->uv1
	sw	t4,28(t1)				;// t4: out_packetp->uv2

	addu	a3,a3,20			;// primtop++
	lw	t2,12(a3)				;// t2: primtop->v0
	lw	t3,16(a3)				;// t3: primtop->v1,v2

	srl	t4,t2,16				;// t4: v0
	sll	t4,t4,3					;// t4: v0*8
	addu	t4,t4,a1			;// t4: (vertop+primtop->v0)

	cfc2	v0,C2_FLAG			;// v0: flag
	nop
	blt	v0,0,TF3L_2				;// clip if flag <0
	nop

			;// CLIP
	li		BITMASK,0x1111

	mfc2	PARAM,C2_SXY0
	jal		CLIP
	swc2    C2_SXY0,8(t1) 		;// sxy0 -->8(pkt_base)

	mfc2	PARAM,C2_SXY1
	jal		CLIP
	swc2    C2_SXY1,16(t1)		;// sxy1 -->16(pkt_base)

	mfc2	PARAM,C2_SXY2
	jal		CLIP
	swc2    C2_SXY2,24(t1)		;// sxy2 -->24(pkt_base)

			;// If any of the 4 bit masks are 1, then there must be 4 points off screen, so clip
	and		BITMASK,0x4444
	bne		BITMASK,zero,TF3L_2
	nop

	;//===========================================================
	NCLIP
	;//===========================================================


	mfc2	v0,C2_MAC0			;// opz read
	beqz	FLIPSTATUS,FLIP_CLIP
	nop
	neg		v0
FLIP_CLIP:
	ble	v0,0,TF3L_2				;// Normal clip if MAC0(OPZ)<=0
	nop

	;//===========================================================
	AVSZ3
	;//===========================================================

	mfc2	t6,C2_OTZ			;// t6: otz
	nop

	srl 	t6,t6,t7			;// sz = sz(otz) >> shift;
	and	t6,t6,0x0ffff			;// mask ?
	sll	t6,t6,2					;// sz*4
	addu	t6,t6,t0			;// t6: &(tagp->org[sz])
	lw	t5,0(t6)				;// t5: table[sz].p

	li	v0,0x7000000			;// (sizeof(POLY_FT3)/4-1)<<24)
	sll	v1,t5,8
	srl	v1,v1,8					;// v1: clear (31..24) of table[sz].p
	or	v0,v0,v1				;// v0: (table[sz].p)|
								;// 	(sizeof(POLY_FT3)/4-1)<<24)
	sw	v0,0(t1)		 		;// *(long *)pkt_base = v0

	nop
	xor	v0,t1,t5				;// v0 : pkt_base xor table[sz].p
	sll	v1,v0,8
	srl	v0,v1,8					;// v1: clear (31..24) of t3
	xor	v1,v0,t5				;// v1: v1 xor table[sz].p
	sw	v1,0(t6)				;// table[sz].p = pkt_base;

	addu	t1,t1,32			;// pkt_base++;

TF3L_2:
	subu t8,t8,1
	bne	t8,0,TF3L_1				;// if(primtop != primtop+n) goto TF3L_1
	nop

TF3L_3:
	lw	v0,0(a0)				;// sp->primp

	opt	at+
	sw	t1,GsOUT_PACKET_P
	opt	at-
	addu v0,v0,8				;// sp->primp+2

		;//Epilogue - Restore the stack
	lw	s0, 0(sp)	// Restore saved args
	lw	s1, 8(sp)
	lw	s2,12(sp)
	lw	s3,16(sp)
	lw	s4,20(sp)
	lw	s5,24(sp)
	lw	ra,28(sp)
	addu	sp,7*4
		;//End of Epilogue

	j       ra			;// return(primtop)
	nop

	opt at+
CLIP:
	sra		SCREEN_Y,PARAM,16				;// signed SY value;
	sll		SCREEN_X,PARAM,16

		;// Check Upper boundary
	bge		SCREEN_Y,CLIP_Y,@SY_OK_UP		;//jump if SCREEN_Y < -128
	sra		SCREEN_X,SCREEN_X,16			;// Sign extend X. (Delay slot)

	addu	BITMASK,BITMASK,0x0001
@SY_OK_UP:
		;// Check left boundary
	bge		SCREEN_X,CLIP_X,@SX_OK_LEFT
	neg		SCREEN_Y

	addu	BITMASK,BITMASK,0x0010
@SX_OK_LEFT:
		;// Check lower boundary
	bgt		SCREEN_Y,CLIP_Y,@SY_OK_DOWN
	neg		SCREEN_X

	addu	BITMASK,BITMASK,0x0100
@SY_OK_DOWN:
		;// Check right boundary
	bgt		SCREEN_X,CLIP_X,@SY_OK_RIGHT
	nop
	addu	BITMASK,BITMASK,0x1000
@SY_OK_RIGHT:
	j		ra
	nop
